@@ -1,6 +1,5 @@
 __author__ = 'Kiril Savino'
 
-import simplejson
 from base import RedisBacked
 
 class Queue(RedisBacked):
@@ -15,6 +14,7 @@ class Queue(RedisBacked):
     >>> import time
 
     >>> q = Queue(Redis('localhost'), 'stuff', ContentType.JSON)
+    >>> q.keep_entry_set = True
     >>> q.clear()
     >>> q.reclaim_tasks()
     >>> assert q.number_active_workers() == 0
@@ -25,6 +25,7 @@ class Queue(RedisBacked):
     >>> string_value = '{"hello": "world"}'
     >>> q.push(dict_value)
     >>> assert q.peek()
+    >>> assert q.contains(dict_value)
     >>> assert q.pop(destructively=True)
     >>> assert not q.pop(destructively=True)
     >>> q.push(string_value)
@@ -38,10 +39,8 @@ class Queue(RedisBacked):
     >>> assert q.number_active_workers() == 1
     >>> q.reclaim_tasks()
 
-    >>> qa = Queue(Redis('localhost'), 'stuff2', ContentType.JSON, worker_id='a')
-    >>> qb = Queue(Redis('localhost'), 'stuff2', ContentType.JSON, worker_id='b')
-    >>> qa.work_ttl_seconds = 1
-    >>> qb.work_ttl_seconds = 1
+    >>> qa = Queue(Redis('localhost'), 'stuff2', ContentType.JSON, worker_id='a', work_ttl=1)
+    >>> qb = Queue(Redis('localhost'), 'stuff2', ContentType.JSON, worker_id='b', work_ttl=1)
     >>> qa.clear()
     >>> qa.push({'hello': 'cruelworld'})
     >>> qa.size()
@@ -74,15 +73,28 @@ class Queue(RedisBacked):
     0
     """
 
-    work_ttl_seconds = 60
+    FIFO = 'fifo'
+    FILO = 'filo'
+    DEFAULT_WORK_TTL_SECONDS = 60
 
-    def __init__(self, redis_client, namespace, content_type, worker_id='global'):
+    def __init__(self, redis_client, namespace, content_type, **kwargs):
+        """
+        optional kwargs:
+        worker_id:       defaults to 'global', but useful if doing multi-processing
+        track_entries:   whether to keep a set around to track membership, defaults to False
+        strategy:        'filo' or 'fifo', defaults to 'fifo'
+        work_ttl:        work_ttl_seconds
+        """
         RedisBacked.__init__(self, redis_client, namespace, content_type)
-        self.worker_id = worker_id
+        self.worker_id = kwargs.get('worker_id', 'global')
         self.QUEUE_LIST_KEY = 'queue.{ns}'.format(ns=namespace)
+        self.ENTRY_SET_KEY = 'queue.{ns}.entries'.format(ns=namespace)
         self.WORKER_SET_KEY = 'queue.{ns}.workers'.format(ns=namespace)
-        self.WORKING_LIST_KEY = 'queue.{ns}.working.{wid}'.format(ns=namespace, wid=worker_id)
-        self.WORKING_ACTIVE_KEY = 'queue.{ns}.active.{wid}'.format(ns=namespace, wid=worker_id)
+        self.WORKING_LIST_KEY = 'queue.{ns}.working.{wid}'.format(ns=namespace, wid=self.worker_id)
+        self.WORKING_ACTIVE_KEY = 'queue.{ns}.active.{wid}'.format(ns=namespace, wid=self.worker_id)
+        self.strategy = kwargs.get('strategy', self.FIFO)
+        self.keep_entry_set = kwargs.get('track_entries', False)
+        self.work_ttl_seconds = kwargs.get('work_ttl', self.DEFAULT_WORK_TTL_SECONDS)
 
     def _on_activity(self):
         self.server.sadd(self.WORKER_SET_KEY, self.worker_id)
@@ -93,6 +105,7 @@ class Queue(RedisBacked):
         with self.server.pipeline() as pipe:
             pipe.multi()
             pipe.delete(self.QUEUE_LIST_KEY)
+            pipe.delete(self.ENTRY_SET_KEY)
             pipe.delete(self.WORKING_LIST_KEY)
             pipe.delete(self.WORKING_ACTIVE_KEY)
             pipe.srem(self.WORKER_SET_KEY, self.worker_id)
@@ -121,7 +134,19 @@ class Queue(RedisBacked):
 
     def push(self, value):
         self._on_activity()
-        self.server.lpush(self.QUEUE_LIST_KEY, self.pack(value))
+        with self.server.pipeline() as pipe:
+            value = self.pack(value)
+            if self.strategy == self.FIFO:
+                pipe.lpush(self.QUEUE_LIST_KEY, value)
+            else:
+                pipe.rpush(self.QUEUE_LIST_KEY, value)
+            if self.keep_entry_set:
+                pipe.sadd(self.ENTRY_SET_KEY, value)
+            pipe.execute()
+
+    def contains(self, value):
+        value = self.pack(value)
+        return self.server.sismember(self.ENTRY_SET_KEY, value)
 
     def pop(self, destructively=False):
         self._on_activity()
@@ -134,7 +159,11 @@ class Queue(RedisBacked):
 
     def complete(self, value):
         self._on_activity()
-        self.server.lrem(self.WORKING_LIST_KEY, self.pack(value))
+        with self.server.pipeline() as pipe:
+            value = self.pack(value)
+            pipe.lrem(self.WORKING_LIST_KEY, value)
+            pipe.srem(self.ENTRY_SET_KEY, value)
+            pipe.execute()
 
     def noop(self):
         self._on_activity()
